/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: config_file.c,v 1.30.2.1.2.22 2003/02/03 23:10:45 lloydy Exp $
 *
 */
#include "config_file.h"
#include "server.h"
#include "chunkyircd.h"

/* Server configuration info */
struct server_config_globals server_config;
struct server_state_globals server_state;

static struct cline **clines = NULL;
static struct nline **nlines = NULL;

static struct tline tlines[8];

/* Command Line Options */
static char *cmdline_conf_file = NULL;

/* Internal Parsing routines */
static int internal_parse_file(char *, struct bind_info[]);
static int internal_motd_load(char *);
static void internal_print_usage();
static void internal_print_version();

/*
* Parse the config file
* Really just a wrapper round internal_parse_file()
*/
int parse_config_file(char *filename,	/* Config File Name */
		      struct bind_info server_sockets[])
{
	time_t curr_time;
	struct tm *split_time;
	char full_path[200];
	int i;

	/* Initialise server_config */
	server_config.conf_file[0] = '\0';
	server_config.olines = NULL;
	server_config.motd = NULL;
	server_config.pass = NULL;
	server_config.services_name[0] = '\0';
	server_config.services_nickserv[0] = '\0';
	server_config.max_fd_buffer_count = 10; /* XXX - Should this be configurable? */

	/* Fill socket array with -1 */
	for (i = 0; i < 32; i++) {
		server_sockets[i].fd = -1;
		server_sockets[i].use_ssl = 0;
	}

	/* NULL out tline array */
	memset(tlines, 0, sizeof(struct tline) * TLINES_MAX);

	/* Set Load Date */
	time(&curr_time);
	split_time = localtime(&curr_time);
	strftime(server_config.load_date, 40, "%a %d %B %Y %H:%M",
		 split_time);

	/* Actually call routine to do parsing */
	if (cmdline_conf_file) {
		/* Use config file spec'ed on cmdline */
		internal_parse_file(cmdline_conf_file, server_sockets);
		strncpy(server_config.conf_file, cmdline_conf_file, 200);
	} else {
		snprintf(full_path, 200, "%s/%s", CONFDIR, filename);
		internal_parse_file(full_path, server_sockets);
		strncpy(server_config.conf_file, full_path, 200);
	}

	/* Free cmdline_* vars */
	if (cmdline_conf_file)
		free(cmdline_conf_file);

	return 1;
}

/* 
* Called to reload the config file
*/
void rehash(int signal)
{
	int i;

	if (signal != SIGHUP) {
		return;
	}

	/* Send Log Message */
	log_msg(LOG_BASIC, "Rehashing from %s", server_config.conf_file);
	log_msg(LOG_BASIC, "Closing log-file");
	close_log();
	open_log(server_config.debug);
	log_msg(LOG_BASIC, "Log reopened, continuing with server Rehash");

	/* Free MOTD */
	for (i = 0;
	     server_config.motd != NULL && server_config.motd[i] != NULL;
	     i++) {
		free(server_config.motd[i]);
		server_config.motd[i] = NULL;
	}
	free(server_config.motd);
	server_config.motd = NULL;

	/* Free cline */
	for (i = 0; clines != NULL && clines[i] != NULL; i++) {
		free(clines[i]);
		clines[i] = NULL;
	}
	free(clines);
	clines = NULL;

	/* Free olines */
	for (i = 0;
	     server_config.olines != NULL
	     && server_config.olines[i] != NULL; i++) {
		free(server_config.olines[i]);
		server_config.olines[i] = NULL;
	}
	free(server_config.olines);
	server_config.olines = NULL;

	/* Free nlines */
	for (i = 0; nlines != NULL && nlines[i] != NULL; i++) {
		free(nlines[i]);
		nlines[i] = NULL;
	}
	free(nlines);
	nlines = NULL;

	/* Clear tlines */
	for (i = 0; i < TLINES_MAX; i++) {
		tlines[i].base.s_addr = 0;
		tlines[i].netmask.s_addr = 0;
	}

	/* Re-initialise modules */
	ircd_module_unload_all();
	ircd_module_rehash();

	/* Reread config file */
	internal_parse_file(server_config.conf_file, NULL);
	/* Add our new server info into struct */
	strncpy(get_my_server()->info, server_config.server_info,
		SERVINFO_MAX);

}

/*
 * Handles the bind node
 */
static void bind_worker(char *key, char *value,
			struct bind_info server_sockets[])
{
	static int curr_socket = 0;
	static struct sockaddr_in bind_ip;
	static int use_ssl = 0;

	/* Only process this node at load time */
	if (server_config.conf_file[0] != '\0') {
		return;
	}

	if (!strcmp(key, "ip")) {
		inet_aton(value, &(bind_ip.sin_addr));
	}

	if (!strcmp(key, "port")) {
		bind_ip.sin_port = htons(atoi(value));
	}

	if (!strcmp(key, "use-ssl")) {
		if (!lc_strncmp(value, "yes", -1)) {
			use_ssl = 1;
		}
	}

	if (!strcmp(key, "}")) {
		/* Bind to it */
		struct sockaddr_in clean;
		int ret_socket;

		/* Allocate and use a clean structure for real bind() */
		memset(&clean, 0, sizeof(struct sockaddr_in));
		clean.sin_family = AF_INET;
		clean.sin_port = bind_ip.sin_port;
		clean.sin_addr.s_addr = bind_ip.sin_addr.s_addr;

		ret_socket = setup_tcp(&clean);

		if (ret_socket != -1) {
			server_sockets[curr_socket].fd = ret_socket;
			server_sockets[curr_socket].use_ssl = use_ssl;
			curr_socket++;
		}

		if (use_ssl) {
			log_msg(LOG_BASIC, "Using SSL on above socket");
		}

		/* Clean up temp store */
		memset(&bind_ip, 0, sizeof(struct sockaddr_in));
		use_ssl = 0;

	}
}

/*
 * Used to parse the main node 
 */
static void main_worker(char *key, char *value,
			struct bind_info server_socket[])
{
	/* Stop the compiler warning */
	server_socket = server_socket;

	/* Only done at load time */
	if (strcmp(key, "name") == 0 && server_config.conf_file[0] == '\0') {
		strncpy(server_config.server_name, value, SERVNAME_MAX);
		return;
	}

	if (!strcmp(key, "info")) {
		strncpy(server_config.server_info, value, SERVINFO_MAX);
		return;
	}

	if (!strcmp(key, "maxchans")) {
		server_config.max_chancount = atoi(value);
		return;
	}

	if (!strcmp(key, "masskicks")) {
		if (lc_strncmp(value, "yes", -1) == 0) {
			server_config.mass_kicks = 1;
		}
		return;
	}

	if (!strcmp(key, "motd")) {
		internal_motd_load(value);
		return;
	}

	if (key[0] == '}') {
		return;
	}

}

/*
 * Handles oper nodes
 */
static void oper_worker(char *key, char *value,
			struct bind_info server_sockets[])
{
	static char pass[20] = "", host[HOST_LENGTH + 1] = "", name[20] =
	    "";
	static int array_size = 0;
	int i;

	server_sockets = server_sockets;

	if (!strcmp(key, "password")) {
		strcpy(pass, value);
	}

	if (!strcmp(key, "hostmask")) {
		strcpy(host, value);
	}

	if (!strcmp(key, "account")) {
		strcpy(name, value);
	}

	if (!strcmp(key, "}")) {
		/* Allocate this nasty stuff */
		if (array_size == 0 || server_config.olines == NULL) {
			server_config.olines =
			    calloc(2, sizeof(struct irc_op *));
			array_size = 2;
		} else {
			array_size++;
			server_config.olines =
			    realloc(server_config.olines,
				    sizeof(struct irc_op *) * array_size);
			server_config.olines[array_size - 1] = NULL;
		}

		for (i = 0; server_config.olines[i] != NULL; i++);

		server_config.olines[i] =
		    ircd_malloc(sizeof(struct irc_op));

		strcpy(server_config.olines[i]->pass, pass);
		strcpy(server_config.olines[i]->name, name);
		strcpy(server_config.olines[i]->hostmask, host);

		pass[0] = '\0';
		name[0] = '\0';
		host[0] = '\0';
	}

	return;
}

/*
* Add a cline
*/
static void connect_worker(char *key, char *value,
			   struct bind_info server_sockets[])
{
	int i;
	static char pass[PASS_LENGTH + 1] = "", host[HOST_LENGTH] =
	    "", name[SERVNAME_MAX] = "";
	static char enc_key[16];
	static int options = 0, port = 0;
	static int array_size = 0;

	/* Stop gcc warning */
	server_sockets = server_sockets;

	if (!strcmp(key, "password")) {
		strncpy(pass, value, PASS_LENGTH);
	}
	if (!strcmp(key, "host")) {
		strncpy(host, value, HOST_LENGTH);
	}
	if (!strcmp(key, "port")) {
		port = atoi(value);
	}
	if (!strcmp(key, "name")) {
		strncpy(name, value, SERVNAME_MAX);
	}
	if (!strcmp(key, "key")) {
		strncpy(enc_key, value, 16);
	}
	if (!strcmp(key, "options")) {
		if (!strcmp(value, "compress")) {
			options |= USR_ZIP;
		}
	}

	if (!strcmp(key, "}")) {


		/* Allocate this nasty stuff */
		if (array_size == 0 || clines == NULL) {
			clines = calloc(2, sizeof(struct cline *));
			array_size = 2;
		} else {
			array_size++;
			clines =
			    realloc(clines,
				    sizeof(struct cline *) * array_size);
			clines[array_size - 1] = NULL;
		}

		for (i = 0; clines[i] != NULL; i++);

		clines[i] = ircd_malloc(sizeof(struct cline));

		strncpy(clines[i]->password, pass, PASS_LENGTH);
		clines[i]->port = port;
		strncpy(clines[i]->host, host, HOST_LENGTH);
		strncpy(clines[i]->serv_name, name, SERVNAME_MAX);
		memcpy(clines[i]->key, enc_key, 16);
	}
	return;
}

/*
 * Add a nline
 */
static void accept_worker(char *key, char *value,
			  struct bind_info server_sockets[])
{
	int i;
	static char pass[PASS_LENGTH + 1] = "", host[HOST_LENGTH] =
	    "", name[SERVNAME_MAX] = "";
	static char enc_key[16];
	static int options = 0;
	static int array_size = 0;

	/* Stop gcc warning */
	server_sockets = server_sockets;

	if (!strcmp(key, "password")) {
		strncpy(pass, value, PASS_LENGTH);
	}
	if (!strcmp(key, "host")) {
		strncpy(host, value, HOST_LENGTH);
	}
	if (!strcmp(key, "name")) {
		strncpy(name, value, SERVNAME_MAX);
	}
	if (!strcmp(key, "key")) {
		strncpy(enc_key, value, 16);
	}
	if (!strcmp(key, "options")) {
		if (!strcmp(value, "compress")) {
			options |= USR_ZIP;
		}
	}

	if (!strcmp(key, "}")) {
		/* Allocate this nasty stuff */
		if (array_size == 0 || nlines == NULL) {
			nlines = calloc(2, sizeof(struct nline *));
			array_size = 2;
		} else {
			array_size++;
			nlines =
			    realloc(nlines,
				    sizeof(struct nline *) * array_size);
			nlines[array_size - 1] = NULL;
		}

		for (i = 0; nlines[i] != NULL; i++);

		nlines[i] = ircd_malloc(sizeof(struct nline));

		strncpy(nlines[i]->password, pass, PASS_LENGTH);
		strncpy(nlines[i]->host, host, HOST_LENGTH);
		strncpy(nlines[i]->serv_name, name, SERVNAME_MAX);
		memcpy(nlines[i]->key, enc_key, 16);
	}
	return;
}

/*
 * Works on the module node
 */
static void modules_worker(char *key, char *value,
			   struct bind_info server_socket[])
{
	/* Avoid warning */
	server_socket = server_socket;

	if (!strcmp(key, "disable")) {
		ircd_module_disable(value);
		return;
	}



	if (!strcmp(key, "}")) {
		/* Apply changes to the module list */
		return;
	}


}

/*
 * Works on the ssl node
 */
static void secure_worker(char *key, char *value,
			  struct bind_info server_socket[])
{
	/* Avoid warning */
	server_socket = server_socket;

#ifdef USE_OPENSSL
	if (!strcmp(key, "cert-file")) {
		strncpy(server_config.ssl_cert_file, value, 200);
		server_config.ssl_cert_file[200] = '\0';
		return;
	}

	if (!strcmp(key, "key-file")) {
		strncpy(server_config.ssl_key_file, value, 200);
		server_config.ssl_key_file[200] = '\0';
		return;
	}
#endif

	/* Add a tline entry */
	if (!strcmp(key, "trusted")) {
		int i;
		char *base_ptr = NULL;
		char *netmask_ptr = NULL;

		base_ptr = value;
		netmask_ptr = value;

		while (*netmask_ptr != '\0') {
			if (*netmask_ptr == '/') {
				*netmask_ptr = '\0';
				*netmask_ptr++;
				break;
			}
			netmask_ptr++;
		}

		log_msg(LOG_DEBUG2,
			"In trusted, got base addr %s and netmask %s",
			base_ptr, netmask_ptr);

		for (i = 0; i < TLINES_MAX; i++) {
			if (tlines[i].base.s_addr == 0) {
				log_msg(LOG_DEBUG2, "Found a space - %d",
					i);
				if (inet_pton
				    (AF_INET, base_ptr,
				     &(tlines[i].base)) <= 0) {
					log_msg(LOG_DEBUG2,
						"Invalid Base Address %s",
						base_ptr);
					/* Clear up what we put in */
					memset(&(tlines[i]), 0,
					       sizeof(struct tline));
					break;
				}
				if (inet_pton
				    (AF_INET, netmask_ptr,
				     &(tlines[i].netmask))
				    <= 0) {
					log_msg(LOG_DEBUG2,
						"Invalid Netmask %s",
						netmask_ptr);
					/* Clear up what we put in */
					memset(&(tlines[i]), 0,
					       sizeof(struct tline));
					break;
				}

				/* Skip out */
				break;
			}
		}

	}


	if (!strcmp(key, "}")) {
		/* Apply changes to the module list */
		return;
	}

}

/*
 * Setup the services config
 */
static void services_worker(char *key, char *value,
			    struct bind_info server_socket[])
{
	server_socket = server_socket;

	if (!strcmp(key, "name")) {
		strncpy(server_config.services_name, value, SERVNAME_MAX);
		server_config.services_name[SERVNAME_MAX] = '\0';
		return;
	}

	if (!strcmp(key, "nickserv")) {
		strncpy(server_config.services_nickserv, value,
			NICK_LENGTH);
		server_config.services_nickserv[NICK_LENGTH] = '\0';
		return;
	}
}

/* Each "Node" is mapped to a particular function */
/* The effect of each node usually isn't applied until the */
/* closing } is reached */
struct config_node nodes[] = {
	{"bind", bind_worker},
	{"main", main_worker},
	{"oper", oper_worker},
	{"connect", connect_worker},
	{"accept", accept_worker},
	{"modules", modules_worker},
	{"secure", secure_worker},
	{"services", services_worker},
	{NULL, NULL},
};


/* 
* This is routine that actually does the parsing
*/
static int internal_parse_file(char *conf_file_name,
			       struct bind_info server_sockets[])
{
	FILE *conf_file;
	char buffer[512];
	char *temp_curr_node = NULL;
	char curr_node[50];
	int i, done;

	conf_file = fopen(conf_file_name, "r");

	if (!conf_file) {
		log_msg(LOG_BASIC, "Failed to open config file %s",
			conf_file_name);
		fprintf(stderr,"Failed to open configuration file %s", conf_file_name);
		exit(1);
	}

	while (fgets(buffer, 511, conf_file)) {
		for (i = 0, done = 0; done == 0 && buffer[i] != '\0'; i++) {
			if (buffer[i] == '{') {
				buffer[i] = '\0';
				done = 1;
			}
		}

		if (done) {
			int keys_done = 0;

			temp_curr_node = buffer;
			for (i = 0; temp_curr_node[i] != '\0'; i++) {
				if (temp_curr_node[i] == ' ') {
					temp_curr_node[i] = '\0';
				}
			}
			strcpy(curr_node, temp_curr_node);

			while (!keys_done) {

				if (fgets(buffer, 511, conf_file)) {
					int done_stripping =
					    0, done_splitting = 0;
					char *stripped = NULL, *value =
					    NULL, *key = NULL;

					/* Strip the tabs and spaces from the start of the string */
					stripped = buffer;
					while (!done_stripping) {

						switch (stripped[0]) {

						case ' ':
						case '\t':
							stripped++;
							break;

						default:
							done_stripping = 1;
							break;
						}
					}

					/* Check for a comment */
					if (stripped[0] == '#') {
						/* Yack! */
						goto break_inner_loop;
					}

					for (i = strlen(stripped) - 1;
					     i >= 0; i--) {
						switch (stripped[i]) {
						case ' ':
						case '\t':
						case '\n':
						case '\r':
							stripped[i] = '\0';
							break;
						default:
							i = -2;
							break;
						}
					}

					/* Parse the stripped text */
					key = stripped;
					while (!done_splitting) {
						switch (stripped[0]) {
						case ' ':
							value =
							    stripped + 1;
							stripped[0] = '\0';
							done_splitting = 1;
							break;
						case '\0':
							value = NULL;
							done_splitting = 1;
							break;
						default:
							stripped++;
							break;
						}

					}

					for (i = 0; nodes[i].name != NULL;
					     i++) {
						if (!strcmp
						    (nodes[i].name,
						     curr_node)) {
							nodes[i].
							    worker(key,
								   value,
								   server_sockets);
						}
					}

					/* If the key is }, break out of this loop */
					if (key[0] == '}') {
						keys_done = 1;
					}

				}
				/* if(fgets()) */
			}	/* while(!keys_done) */
		      break_inner_loop:
			;

		}		/* if(done) */
	}			/* while(fgets) */

	fclose(conf_file);
	return 1;
}

/*
* Load the motd 
*/
static int internal_motd_load(char *filename)
{

	int motd_fd = -1;
	long int curr_offset = 0;
	int line_len = 0;
	char buffer[1024];
	int retval = 0;
	int array_size = 0;
	int i, j = 0;

	motd_fd = open(filename, O_RDONLY);

	if (motd_fd == -1) {
		log_msg(LOG_BASIC, "Failed to open MOTD file %s",
			filename);
		return 0;
	}


	while ((retval = read(motd_fd, buffer, 1024)) > 0) {
		buffer[retval] = 0;

		for (i = 0; buffer[i] != '\0' && line_len == 0; i++) {
			if (buffer[i] == '\n') {
				line_len = i + 1;
				buffer[i] = '\0';
			}
		}

		if (server_config.motd == NULL) {
			server_config.motd = calloc(2, sizeof(char *));
			server_config.motd[0] = NULL;
			server_config.motd[1] = NULL;
			array_size = 2;
		} else {
			array_size++;
			server_config.motd =
			    realloc(server_config.motd,
				    array_size * sizeof(char *));
			server_config.motd[array_size - 1] = NULL;
		}

		server_config.motd[j] = string_alloc(strlen(buffer));
		strcpy(server_config.motd[j], buffer);

		j++;
		curr_offset += line_len;
		lseek(motd_fd, curr_offset, SEEK_SET);
		line_len = 0;
	}

	close(motd_fd);
	return 1;
}


#ifdef HAVE_GETOPT_H
/* Only valid with GNU getopt */
/* Command line params are in this array of structs */
struct option long_options[] = { {"debug", 1, 0, 'd'},
{"help", 0, 0, 'h'},
{"config-file", 1, 0, 'c'},
{"foreground", 0, 0, 'f'},
{"version", 0, 0, 'V'},
{NULL, 0, 0, 0},
};
#endif

/* 
* Command line parser
*/
int parse_cmdline(int argc, char *argv[])
{

	char retval;

#ifdef HAVE_GETOPT_H
	int option_index;

	while ((retval =
		getopt_long(argc, argv, "i:p:d:fhVc:", long_options,
			    &option_index)) != EOF) {
#else
	while ((retval = getopt(argc, argv, "i:p:d:fhVc:")) != EOF) {
#endif

		switch (retval) {

		case 'd':
			server_config.debug = atoi(optarg);
			break;

		case 'c':
			cmdline_conf_file = string_alloc(strlen(optarg));
			strcpy(cmdline_conf_file, optarg);
			break;

		case 'f':
			server_config.foreground = 1;
			break;

		case 'V':
			internal_print_version();
			exit(0);
			break;

		case 'h':
			internal_print_usage();
			exit(0);
			break;
		}

	}

	return 1;
}

/*
* Get a CLINE when we try and connect 
*/
struct cline *get_cline(char *serv_name)
{
	int i, done = -1;

	if (clines == NULL) {
		return NULL;
	}

	for (i = 0; clines[i] != NULL && done == -1; i++) {
		if ((lc_strncmp(clines[i]->serv_name, serv_name, -1) == 0)) {
			done = i;
		}
	}

	if (done != -1) {
		return clines[done];
	} else {
		return NULL;
	}
}

/*
* Get a NLINE when we try and connect 
*/
struct nline *get_nline(char *host)
{
	int i, done = -1;

	if (nlines == NULL) {
		return NULL;
	}

	for (i = 0; nlines[i] != NULL && done == -1; i++) {
		if ((lc_strncmp(nlines[i]->host, host, -1) == 0)) {
			done = i;
		}
	}

	if (done != -1) {
		return nlines[done];
	} else {
		return NULL;
	}
}

/*
 * Check a tline
 * Return 0 if no match found, or -1 if a match
 */
int check_tlines(struct in_addr *ip_to_check)
{
	int i;
	char buf[32];
	char buf1[32];
	char buf2[32];

	for (i = 0; i < TLINES_MAX && tlines[i].base.s_addr != 0; i++) {
		if ((tlines[i].base.s_addr & tlines[i].netmask.s_addr)
		    == (ip_to_check->s_addr & tlines[i].netmask.s_addr)) {
			log_msg(LOG_DEBUG2,
				"Trusting IP %s on network %s with netmask %s",
				inet_ntop(AF_INET, ip_to_check, buf, 30),
				inet_ntop(AF_INET, &(tlines[i].base), buf1,
					  30), inet_ntop(AF_INET,
							 &(tlines[i].
							   netmask), buf2,
							 30));
			return -1;
		}
	}

	return 0;

}

/*
* Print usafe
*/
static void internal_print_usage()
{
	printf("Chunky Monkey IRCD Version %s\n", VERSION);
	printf("Copyright (c) Chris Plant 2000-2001\n");
	printf("	      David Leadbeater 2001\n");
	printf("	      Mike Saunders 2001\n");
	printf("\n");
	printf("Usage: ircd\n");
#ifdef HAVE_GETOPT_H
	printf(" -c, --config-file=	Config File to use\n");
	printf(" -d, --debug=		Set debug level\n");
	printf(" -f, --foreground	Don't turn into a daemon\n");
	printf(" -h, --help 		Print this message\n");
	printf(" -V, --version		Print version\n");
#else
	printf(" -c <config_file>	Config File to use\n");
	printf(" -d <debug level>	Set debug level\n");
	printf(" -f			Don't turn into a daemon\n");
	printf(" -h	 		Print this message\n");
	printf(" -V			Print version\n");
#endif
}

/*
 * Print version and GNU license summary
 */
static void internal_print_version()
{
	printf("Chunky Monkey IRCD Version %s\n", VERSION);

}
