/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: resolver.c,v 1.10.4.6 2003/01/19 12:17:14 lloydy Exp $
 *
 */
#include "resolver.h"
#include <errno.h>
#include "chunkyircd.h"

/* State of adns system */
adns_state ads;

/* Linked list start */
static LookupInfo list_start = NULL;


/*
 * Add an item into the resolver linked list
 */
static int resolver_list_add(LookupInfo lookup)
{
	LookupInfo curr = NULL, prev = NULL;

	if (!list_start) {
		lookup->next = NULL;
		lookup->id = 1;
		list_start = lookup;
	} else {
		curr = list_start;
		while (curr
		       && (!curr->next
			   || curr->id + 1 == curr->next->id)) {
			prev = curr;
			curr = curr->next;
		}
		if (!curr) {
			lookup->next = prev->next;
			prev->next = lookup;
			lookup->id = prev->id + 1;
		} else {
			lookup->next = curr->next;
			curr->next = lookup;
		}
	}

	return lookup->id;
}

/*
 * Remove an item from the linked list
 */
static int resolver_list_remove(int id)
{
	LookupInfo curr = NULL, prev = NULL;

	if (!list_start) {
		return 0;
	} else {
		curr = list_start;
		while (curr && curr->id != id) {
			prev = curr;
			curr = curr->next;
		}
		if (curr && curr->id == id) {
			if (prev) {
				prev->next = curr->next;
			} else {
				list_start = curr->next;
			}
			free(curr);
		}

	}

	return 1;
}

/*
 * Get a LookupInfo for a query
 */
static LookupInfo resolver_list_get(int id)
{
	LookupInfo curr = NULL, prev = NULL;

	if (!list_start) {
		return NULL;
	} else {
		curr = list_start;
		while (curr && curr->id != id) {
			prev = curr;
			curr = curr->next;
		}
		if (curr && curr->id == id) {
			return curr;
		} else {
			return NULL;
		}
	}
}

/* 
 * Setup adns for stuff 
 */
int resolver_init()
{
	adns_initflags initflags;
	int retval;

	if (server_config.debug > 1) {
		initflags = adns_if_debug;
	} else {
		initflags = 0;
	}

	retval = adns_init(&ads, initflags, 0);

	if (retval) {
		log_msg(LOG_BASIC, "Failed to initialise adns");
		return 0;
	}


	return 1;
}

/*
 * Submit an IP for reverse lookup to the adns library
 */
int resolver_submit_reverse(unsigned int ip)
{
	char str[30];
	unsigned int temp1 = 0, temp2 = 0, temp3 = 0, temp4 = 0;
	LookupInfo lookup = NULL;

	lookup = ircd_malloc(sizeof(struct _lookup));

	lookup->ip = ip;

	lookup->flags = RES_REVERSE;

	/* Add in */
	temp1 = ip >> 24;
	temp2 = (ip - (temp1 << 24)) >> 16;
	temp3 = (ip - (temp1 << 24) - (temp2 << 16)) >> 8;
	temp4 = ip - (temp1 << 24) - (temp2 << 16) - (temp3 << 8);

	/* Submit to adns */
	sprintf(str, "%0d.%0d.%0d.%0d.in-addr.arpa.", temp1, temp2, temp3,
		temp4);
	adns_submit(ads, str, adns_r_ptr,
		    adns_qf_quoteok_cname | adns_qf_cname_loose, NULL,
		    &(lookup->query));



	return resolver_list_add(lookup);



}

/* 
 * Submit a non-reverse lookup
 */
int resolver_submit(char *hostname)
{
	char buffer[HOST_LENGTH + 2];
	int length = 0;
	int retval;
	LookupInfo lookup = NULL;

	lookup = ircd_malloc(sizeof(struct _lookup));

	/* Check for a . at the end */
	length = strlen(hostname);
	if (hostname[length - 1] == '.') {
		strncpy(buffer, hostname, HOST_LENGTH);
	} else {
		ircd_sprintf(buffer, HOST_LENGTH + 2, "%s.", hostname);
	}

	retval = adns_submit(ads, buffer, adns_r_a,
			     adns_qf_quoteok_cname | adns_qf_cname_loose,
			     NULL, &(lookup->query));

	if (retval == ENOSYS) {
		log_msg(LOG_BASIC,
			"Adns is shitting its pants in %s on line %d\n",
			__FILE__, __LINE__);
	}

	return resolver_list_add(lookup);
}

/* 
 * Check the state of an item 
 */
char *resolver_check(int id)
{
	adns_answer *reply;
	int retval;
	static char hostname[HOST_LENGTH + 1];
	struct in_addr temp;
	LookupInfo lookup = NULL;

	if (id == -1) {
		return NULL;
	}

	lookup = resolver_list_get(id);

	if (!lookup) {
		return NULL;
	}

	/* Check for ip = 127.0.0.1 */
	inet_aton("127.0.0.1", &temp);
	if (lookup->ip == temp.s_addr) {
		return get_my_server()->name;
	}

	retval = adns_check(ads, &(lookup->query), &reply, NULL);

	if (retval == EAGAIN) {
		return NULL;
	}

	if (retval) {
		log_msg(LOG_BASIC, "RESOLVER: Error: %s\n",
			strerror(retval));
		return NULL;
	}

	if (reply->status == adns_s_ok) {
		if (lookup->flags & RES_REVERSE) {
			/* Its a reverse lookup */
			strncpy(hostname, *reply->rrs.str, HOST_LENGTH);
		} else {
			/* Its a normal lookup */
			strncpy(hostname, inet_ntoa(*reply->rrs.inaddr),
				HOST_LENGTH);
		}
	} else {
		/* Lookup failed, just whack the IP in */
		temp.s_addr = lookup->ip;
		strncpy(hostname, (char *) inet_ntoa(temp), HOST_LENGTH);
	}
	resolver_list_remove(id);
	return hostname;

}

/* 
 * Cancel a resolve request from adns
 * and remove it from the list
 */
int resolver_cancel(int id)
{
	LookupInfo lookup = NULL;

	if (id == -1)
		return 0;

	lookup = resolver_list_get(id);

	if (lookup) {
		adns_cancel(lookup->query);
		resolver_list_remove(id);
		return 1;
	} else {
		return 0;
	}
}
