/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: temp_server.c,v 1.18.4.17 2003/02/04 23:31:50 lloydy Exp $
 *
 */
#include "temp_server.h"
#include "ircd.h"
#include "enpress.h"

/* temp_server.c - Routines to handle a SERVER command from a 
 * peer in the temporary list */

int temp_server(struct tempinfo *temp, Cmd cmd)
{
	/* params[0] is the server name
	 * params[1] is the hopcount, 0 in this case
	 * params[2] is the token, not sure we need it
	 * text is the server info field */
	struct nline *nline = NULL;

	/* Check that this peer has successfully used PASS */
	if (!(temp->flags & USR_PASS)) {
		force_quit(temp->fd,
			   "You are trying to establish a server link from an unassured host");
		return 0;
	}


	if (!check_params(cmd, 1)) {
		return 0;
	}

	if (cmd->text) {
		/* Check for duplicates */
		if (server_locate(cmd->params[0])) {
			/* Dup server */
			ircd_fdlist_clear_entry(temp->fd);
			temp_rem(temp->fd);
			close(temp->fd);
			return 0;
		}

		/* Check against nline */
		nline = get_nline(temp->host);
		if (nline == NULL
		    || lc_strncmp(nline->serv_name, cmd->params[0],
				  -1) != 0) {
			force_quit(temp->fd, "Your host isn't authorised");
			return 0;
		}

		strncpy(temp->info, cmd->text, SERVINFO_MAX);
		strncpy(temp->name, cmd->params[0], SERVNAME_MAX);
		temp->flags |= USR_SERVER;
		send_temp(temp, "CAPAB :TS4");
	}
	return 1;
}

/*
* Deal with PASS in temp state
*/
int temp_pass(struct tempinfo *t_ptr, Cmd cmd)
{
	/* Format */
	/* PASS <password> <version> <flags> [other stuff] */
	/* We don't have any options/flags atm :) */
	/* So we just need to check the password and the version is 0.94d+ cmircd */
	char *pass = NULL;
	struct nline *nline = NULL;
	size_t p_len;
	size_t act_len;

	if (!(cmd->params[0]) && !(cmd->text)) {
		return 0;
	}

	if (cmd->params[0]) {
		pass = cmd->params[0];
	} else {
		pass = cmd->text;
	}

	if ( !check_params(cmd, 3) ) {
		/* Client Password */
		if (!server_config.pass) {
			/* No Local Password set, forward to services */
			if (server_config.services_name[0] != '\0' &&
			    server_config.services_nickserv[0] != '\0' &&
			 	t_ptr->pass == NULL) {
					
				/* Store, ready to forward */
				p_len = strlen(pass);

				act_len = p_len > 30?30:p_len;
				
				t_ptr->pass = string_alloc(act_len);

				strncpy(t_ptr->pass, pass, act_len);
				t_ptr->pass[act_len] = '\0';

			}
		}
	} else {

		/* Get the NLINE for this host */
		nline = get_nline(t_ptr->host);

		if (nline) {
			if (strcmp(nline->password, cmd->params[0]) == 0) {
				t_ptr->flags |= USR_PASS;
			}
		}
	}

	return 1;
}


/*
* Deal with CAPAB required by the TS4
* Protocol
*/
int temp_capab(struct tempinfo *t_ptr, Cmd cmd)
{
	int i, last, curr = 0, len;
	char temp[300];
	char *capabs[4] = { NULL, NULL, NULL, NULL };
#ifdef USE_OPENSSL
	EncryptInfo ctx = NULL;
	struct nline *n_ptr = NULL;
#endif

	/* We need cmd->text */
	if (!cmd->text) {
		return 0;
	}

	/* Split it up, set spaces to \0 */
	len = strlen(cmd->text);
	strncpy(temp, cmd->text, (len > 300)?300:len );
	temp[(len > 300)?300:len] = '\0';
	last = 0;

	for (i = 0; i <= len && curr < 4; i++) {
		if (temp[i] == ' ' || temp[i] == '\0') {
			capabs[curr++] = &(temp[last]);
			temp[i] = '\0';
			i++;
			last = i;
		}
	}

	for (i = 0; capabs[i] != NULL; i++) {
		if (lc_strncmp(capabs[i], "ZIP", -1) == 0) {
			/* Set SERVER_ZIP flag */
			t_ptr->flags |= USR_ZIP;
		}
		if (lc_strncmp(capabs[i], "TS4", -1) == 0) {
			/* Set SERVER_TS4 flags */
			t_ptr->flags |= USR_TS4;
		}
#ifdef USE_OPENSSL
		if (lc_strncmp(capabs[i], "ENCRYPT", -1) == 0) {
			/* Don't repeat this if we're already using encryption */
			if (!(t_ptr->enc_ctx)) {
				/* Server wants encryption, use key from the file */
				/* Get NLINE for the key */
				n_ptr = get_nline(t_ptr->host);
				/* Try and get context to encrypt */
				ctx =
				    generate_key(n_ptr ? n_ptr->
						 key : (unsigned char *)
						 "*");
				if (ctx) {
					log_msg(LOG_BASIC,
						"GOT CAPAB :ENCRYPT, going encrypted");
					/* we can go encrypted */
					t_ptr->enc_ctx = ctx;
					send_temp(t_ptr, "CAPAB :ENCRYPT");

				} else {
					/* No context to encrypt with
					 * drop peer */
					force_quit(t_ptr->fd,
						   "Attempt to use encryption without local key");
					/* No point in continuation of processing without key */
					return -1;
				}
			}

		}
#endif
	}

	return 1;
}
